import sys

from madmigration.config.conf import ConfigYamlManager
from madmigration.db_operations.operations import DBOperations
from madmigration import basemigration
from madmigration.utils.logger import configure_logging

logger = configure_logging(__file__)

class Controller:
    def __init__(self, config: ConfigYamlManager):
        self.config = config
        self.source_db_operations = DBOperations(self.config.source_uri)
        self.destination_db_operations = DBOperations(self.config.destination_uri, create_database=True)

    def __enter__(self):
        return self


    def __exit__(self, *args):
        self.source_db_operations.session.close()
        self.destination_db_operations.session.close()


    def sig_handler(self, sig_num, _sig_frame):
        logger.warn(f"Terminate app with signal -> {sig_num}")
        self.destination_db_operations.db_drop_everything()
        sys.exit(sig_num)
        

    def get_base_migration_class_with_database_driver_name(self, driver):
        return {
            "mysqldb": basemigration.MysqlMigrate,
            "mysql+mysqldb": basemigration.MysqlMigrate,
            "pymysql": basemigration.MysqlMigrate,
            "mysql+pymysql": basemigration.MysqlMigrate,
            "mariadb+pymsql": basemigration.MysqlMigrate,
            "psycopg2": basemigration.PgMigrate,
            "pg8000": basemigration.PgMigrate,
            # "pyodbc": MssqlMigrate,
            # "mongodb": MongoDbMigrate
        }.get(driver)

    def run(self):
        self.destination_db_driver_name = self.destination_db_operations.engine.driver
        migration = self.get_base_migration_class_with_database_driver_name(self.destination_db_driver_name)
        mig = migration(self.config, self.source_db_operations, self.destination_db_operations)
        mig.run()


class NoSQLControllerWithYamlConf(Controller):
    pass


class SqlControllerWithYamlConf(Controller):
    def __init__(self, config: ConfigYamlManager):
        super().__init__(config)



class FullMigrateController(Controller):
    def __init__(self, config: ConfigYamlManager):
        super().__init__(config)

    def run():
        pass



def run(config_file, full_migrate):
    conf = ConfigYamlManager(config_file)
    
    if conf.destination_mongo:
        controller = NoSQLControllerWithYamlConf
    elif not conf.destination_mongo:
        controller = SqlControllerWithYamlConf
    elif full_migrate:
        controller = FullMigrateController

    with controller(conf) as c:
        c.run()



import signal
import sys
from queue import Queue

from madmigration.config.config_schema import MigrationTablesSchema
from madmigration.config.config_schema import ColumnParametersSchema
from sqlalchemy import Column, MetaData
from sqlalchemy.engine import reflection
from collections import defaultdict
from madmigration.postgresqldb.type_convert import get_type_object
from madmigration.config.conf import ConfigYamlManager
from madmigration.db_operations.operations import DbOperations
from madmigration.utils.logger import configure_logging


logger = configure_logging(__name__)


class BaseMigrate():
    q = Queue()  # Static queue for fk constraints data
    tables = set()

    def __init__(self, config: ConfigYamlManager, destination_db):
        self.config = config
        self.migration_tables = self.config.migrationTables
        self.engine = destination_db.engine
        self.connection = destination_db
        self.metadata = MetaData()
        self.table_list = set()
        self.table_create = defaultdict(list)
        self.table_update = defaultdict(list)
        self.alter_col = defaultdict(list)
        self.fk_constraints = []
        self.dest_fk = []
        self.contraints_columns = defaultdict(set)
        self.db_operations = DbOperations(self.engine)
        self._drop_tables = []

        signal.signal(signal.SIGINT, self.sig_handler)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.engine.session.close()

    def sig_handler(self, sig_num, _sig_frame):
        logger.warn("TERMINATE APP WITH SIGNAL -> %d" % sig_num)
        if self.tables:
            self.db_operations.db_drop_everything(self.tables)
        sys.exit(sig_num)

    def collect_table_names(self):
        """Collects all tables that the program should create"""
        self.tables = self.config.collect_destination_tables()

    def parse_migration_tables(self, tabels_schema: MigrationTablesSchema):
        """
        This function parses migrationTables from yaml file
        """
        try:
            self.source_table = tabels_schema.migrationTable.SourceTable.dict()
            self.destination_table = tabels_schema.migrationTable.DestinationTable.dict()
            self.columns = tabels_schema.migrationTable.MigrationColumns
        except Exception as err:
            logger.error("parse_migration_tables [error] -> %s" % err)

    def parse_migration_columns(
        self, tablename: str, migration_columns: ColumnParametersSchema
    ):
        """
        This function parses migrationColumns schema and prepare column
        """
        try:
            table_exists = self.db_operations.is_table_exists(tablename)

            for col in migration_columns:
                self.source_column = col.sourceColumn
                self.destination_column = col.destinationColumn
                self.dest_options = col.destinationColumn.options.dict()

                self._parse_fk(tablename, self.dest_options.pop("foreign_key"))
                column_type = self._parse_column_type()

                col = Column(self.destination_column.name, column_type, **self.dest_options)
                if table_exists:
                    if not self.db_operations.is_column_exists_in_table(tablename, self.destination_column.name):
                        #     self.add_alter_column(tablename, {"column_name": self.destination_column.name,"type":column_type,"options":{**self.dest_options}})
                        # else:
                        self.add_updated_table(tablename, col)
                else:
                    self.add_created_table(tablename, col)
        except Exception as err:
            logger.error("parse_migration_columns [error] -> %s" % err)

    def add_updated_table(self, table_name: str, col: Column):
        self.table_update[table_name].append(col)

    def add_created_table(self, table_name: str, col: Column):
        self.table_create[table_name].append(col)

    def add_alter_column(self, table_name: str, col: Column):
        self.alter_col[table_name].append(col)

    def prepare_tables(self):
        try:
            for migrate_table in self.migration_tables:
                if migrate_table.migrationTable.DestinationTable.create:
                    self.parse_migration_tables(migrate_table)
                    self.parse_migration_columns(self.destination_table.get("name"), self.columns)
        except Exception as err:
            logger.error("prepare_tables  [error] -> %s" % err)

    def update_table(self):

        for tab, col in self.table_update.items():
            self.db_operations.add_column(tab, *col)
        return True

    def alter_columns(self):
        for tab, val in self.alter_col.items():
            for i in val:
                self.db_operations.update_column(tab, i.pop("column_name"), i.pop("type"), **i.pop("options"))
        return True

    def create_tables(self):
        for tab, col in self.table_create.items():
            self.db_operations.create_table(tab, *col)
        return True

    def process(self):
        """
        Create and check existing tables. 
        Collect foreign key constraints
        """
        try:
            self.dest_fk, self.contraints_columns = self.db_operations.collect_fk_and_constraint_columns(self.table_list)
            if self._drop_tables:
                self.db_operations.bulk_drop_tables(*self._drop_tables)
            self.update_table()
            self.create_tables()
            self.db_operations.create_fk_constraint(self.fk_constraints,self.contraints_columns)
            return True
        except Exception as err:
            logger.error("create_tables [error] -> %s" % err)

    def _parse_column_type(self) -> object:
        """ Parse column type and options (length,type and etc.) """

        try:
            column_type = self.get_column_type(self.dest_options.pop("type_cast"))
            type_length = self.dest_options.pop("length")
            if type_length:
                column_type = column_type(type_length)
            return column_type
        except Exception as err:
            logger.error("_parse_column_type [error] -> %s" % err)

        # logger.error(self.dest_options.get("length"))
        type_length = self.dest_options.pop("length")
        if type_length:
            column_type = column_type(type_length)
        return column_type

    def _parse_fk(self, tablename, fk_options):
        """ Parse foreignkey and options (use_alter,colum and etc.) """
        try:
            if fk_options:
                fk_options["source_table"] = tablename
                fk_options["dest_column"] = self.destination_column.name
                self.fk_constraints.append(fk_options)
        except Exception as err:
            logger.error("_parse_fk [error] -> %s" % err)

    def get_input(self,table_name):
        while True:
            answ = input(
                f"Table with name '{table_name}' already exist,\
'{table_name}' table will be dropped and recreated,your table data will be lost,process?(y/n) ")
            if answ.lower() == "y":
                if self.dest_fk[table_name]:
                    self.db_operations.drop_fk(self.dest_fk[table_name])
                self._drop_tables.append(table_name)
                return False
            elif answ.lower() == "n":
                return True
            else:
                continue

    def get_table_attribute_from_base_class(self, source_table_name: str):
        """
        This function gets table name attribute from sourceDB.base.classes. Example sourceDB.base.class.(table name)
        Using this attribute we can query table using sourceDB.session
        :return table attribute
        """
        return getattr(self.connection.base.classes, source_table_name)

    def get_data_from_source_table(self, source_table_name: str, source_columns: list):

        table = self.get_table_attribute_from_base_class(source_table_name.name)
        rows = self.connection.session.query(table).yield_per(1)

        for row in rows:
            data = {}
            for column in source_columns:
                data[column] = getattr(row, column)
            yield data

    @staticmethod
    def insert_data(engine, table_name, data: dict):
        # stmt = None
        try:
            stmt = engine.base.metadata.tables[table_name].insert().values(**data)

        except Exception as err:
            logger.error("insert_data stmt [error] -> %s" % err)
            return
        # logger.error("STMT",stmt)

        try:
            engine.session.execute(stmt)
        except Exception as err:
            logger.error("insert_data passing into queue [error] -> %s" % err)
            BaseMigrate.put_queue(stmt)

        try:
            engine.session.commit()
        except Exception as err:
            logger.error("insert_data [error] -> %s" % err)
            engine.session.rollback()
        finally:
            engine.session.close()

    @staticmethod
    def insert_queue(engine):
        for stmt in BaseMigrate.q.queue:

            try:
                logger.info("Inserting from queue")
                engine.session.execute(stmt)
            except Exception as err:
                logger.error("insert_queue [error] -> %s" % err)

            try:
                engine.session.commit()
            except Exception as err:
                logger.error("insert_queue [error] -> %s" % err)
                engine.session.rollback()
            finally:
                engine.session.close()

    @staticmethod
    def put_queue(data):
        BaseMigrate.q.put(data)

    @staticmethod
    def get_queue():
        return BaseMigrate.q.get()

    @staticmethod
    def type_cast(data_from_source, mt, convert_info: dict):
        for columns in mt.migrationTable.MigrationColumns:
            source_column = columns.sourceColumn.name
            destination_column = columns.destinationColumn.name

            if columns.destinationColumn.options.type_cast:
                destination_type_cast = columns.destinationColumn.options.type_cast
            else:
                destination_type_cast = None

            if convert_info.get(destination_column):
                # ClassType is Class of data type (int, str, float, etc...)
                # Using this ClassType we are converting data into format specified in type_cast
                datatype = get_type_object(destination_type_cast)

                try:
                    if datatype == type(data_from_source.get(source_column)):
                        data_from_source[destination_column] = data_from_source.pop(source_column)

                    elif datatype.__name__ == "UUID":
                        data_from_source[destination_column] = str(data_from_source.pop(source_column))
                    else:
                        try:
                            data_from_source[destination_column] = datatype(data_from_source.pop(source_column))
                        except Exception as err:
                            logger.error("type_cast [error] -> %s" % err)
                            data_from_source[destination_column] = None

                except Exception as err:
                    logger.error("type_cast [error] -> %s" % err)
                    data_from_source[destination_column] = None
            else:
                data_from_source[destination_column] = data_from_source.pop(source_column)

        return data_from_source

    @staticmethod
    def get_column_type(type_name: str) -> object:
        """Get class of db type
        :param type_name: str
        :return: object class
        """
        raise NotImplementedError



from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.dialects.postgresql import (
    ARRAY,
    BIGINT,
    BIT,
    BOOLEAN,
    BYTEA,
    CHAR,
    DATE,
    # DOUBLE_PRECISION,
    # ENUM,
    FLOAT,
    INET,
    INTEGER,
    # INTERVAL,
    JSON,
    JSONB,
    MACADDR,
    MONEY,
    NUMERIC,
    OID,
    REAL,
    SMALLINT,
    TEXT,
    TIME,
    TIMESTAMP,
    VARCHAR,
)
from sqlalchemy import DateTime
from sqlalchemy_utils import UUIDType
from madmigration.basemigration.base_migration import BaseMigrate
from madmigration.config.conf import ConfigYamlManager


class PgMigrate(BaseMigrate): 
    def __init__(self, config: ConfigYamlManager, source_db_operations, destination_db_operations):
        super().__init__(config, source_db_operations, destination_db_operations)

    @staticmethod
    def get_column_type(type_name: str) -> object:
        """Get class of db type"""
        return {
            "varchar": VARCHAR,
            "char": CHAR,
            "string": VARCHAR,
            "text": TEXT,
            "integer": INTEGER,
            "smallint": SMALLINT,
            "bigint": BIGINT,
            "binary": BYTEA,
            "boolean": BOOLEAN,
            "bool": BOOLEAN,
            "date": DATE,
            "datetime": DateTime,
            "timestamp": TIMESTAMP,
            "time": TIME,
            # "enum": ENUM,
            "float": FLOAT,
            "real": REAL,
            "json": MutableDict.as_mutable(JSON),
            "jsonb": MutableDict.as_mutable(JSONB),
            # "array": ARRAY, #FIXME column with array include array elemet type argument
            "numeric": NUMERIC,
            "money": MONEY,
            "macaddr": MACADDR,
            "inet": INET,
            "oid": OID,
            "uuid": UUIDType(binary=False),
        }.get(type_name.lower())





from pydantic import BaseModel, validator
from typing import Union, List, Any, AnyStr, Dict, Optional


class ForeignKeySchema(BaseModel):
    use_alter: bool = True
    table_name: str
    column_name: str
    ondelete: str = "CASCADE"
    onupdate: str = "NO ACTION"


class OptionsSchema(BaseModel):
    primary_key: bool = False
    nullable: bool = False
    default: Any = None
    index: Union[bool, None] = None
    unique: Union[bool, None] = None
    autoincrement: bool = False
    foreign_key: Union[ForeignKeySchema, None] = None
    length: Union[int, None] = None
    type_cast: Union[str, None] = None

    # @validator("type_cast")   #TODO check types
    # def validate_cast(cls,v,values):
    #     if v == "integer" and values["length"] != None:
    #         raise ValueError(f"Type {v} has no length")
    #     return v


class SourceConfigSchema(BaseModel):
    SourceConfig: Dict[str, str]


class DestinationConfigSchema(BaseModel):
    DestinationConfig: Dict[str, str]


class SourceTableSchema(BaseModel):
    name: str


class DestTableSchema(BaseModel):
    name: str
    create: bool = False


class DestinationColumn(BaseModel):
    name: str
    options: OptionsSchema


class SourceColumn(BaseModel):
    name: str


class ColumnParametersSchema(BaseModel):
    destinationColumn: DestinationColumn
    sourceColumn: SourceColumn


class TablesInfo(BaseModel):
    SourceTable: Optional[SourceTableSchema]
    DestinationTable: Optional[DestTableSchema]
    MigrationColumns: Optional[List[ColumnParametersSchema]]


class MigrationTablesSchema(BaseModel):
    migrationTable: Optional[TablesInfo]


class ConfigSchema(BaseModel):
    Configs: List[Union[SourceConfigSchema, DestinationConfigSchema]]
    migrationTables: Optional[List[MigrationTablesSchema]]
    version: float


class OptionsSchmea(BaseModel):
    options: Dict[str, str]


class ConcatenateColumns(ColumnParametersSchema):

    destinationColumn: OptionsSchmea


class Options(BaseModel):
    type_cast: str
    concatenateColumns: List[ConcatenateColumns] = None
    concatenateTable: str = None


class DestinationColumnSchema(DestinationColumn):
    options: Options


class ColumConfig(ColumnParametersSchema):
    destinationColumn: DestinationColumnSchema  # options dict


class DestinationTableSchema(SourceTableSchema):
    ...


class TablesSchema(TablesInfo):
    DestinationTable: DestinationTableSchema
    MigrationColumns: List[ColumConfig]


class MigrationTableS(MigrationTablesSchema):
    migrationTable: TablesSchema


class NoSQLConfigSchema(ConfigSchema):

    migrationTables: List[MigrationTableS]




from sqlalchemy import create_engine, event, Table, MetaData, ForeignKeyConstraint
from sqlalchemy.schema import DropConstraint, DropTable
from sqlalchemy.orm import Session
from sqlalchemy.ext.automap import automap_base
from sqlalchemy_utils.functions.database import database_exists, create_database
from sqlalchemy.engine.url import make_url
from sqlalchemy.ext.declarative import declarative_base
from madmigration.utils.logger import configure_logging
from madmigration.utils.helpers import goodby_message, database_not_exists
from alembic.migration import MigrationContext
from sqlalchemy.engine import reflection
from alembic.operations import Operations
from contextlib import contextmanager
import sys

logger = configure_logging(__name__)


 
@contextmanager
def OperationContextManager(engine):
    try:
        conn = engine.connect()
        ctx = MigrationContext.configure(conn)
        op = Operations(ctx)
        yield op
    except Exception as err:
        logger.error("OperationContextManager ->", err)
        sys.exit(1)
    finally:
        conn.close()

@contextmanager
def InspectorReflection(engine):
    try:
        inspector = reflection.Inspector.from_engine(engine)
        yield inspector
    except Exception as err:
        print('err ->', err)
        logger.error('InspectorReflection ->', err, exc_info=True)
        sys.exit(1)


@contextmanager
def Transaction(engine):
    try:
        conn = engine.connect()
        transactional = conn.begin()
        yield transactional
    except Exception as err:
        logger.error('Transaction ->', err)
        sys.exit(1)
    finally:
        conn.close()


class DBOperations:
    def __init__(self, uri, create_database=False):
        try:
            self.uri = uri
            self.password_masked_database_uri = make_url(self.uri)

            self.engine = create_engine(uri, echo=False)
            self.session = Session(self.engine, autocommit=False, autoflush=False)
            
            self.metadata = MetaData(bind=self.engine)

            if create_database:
                self.create_database_if_not_exists()

            self.base = declarative_base()
            self.base.metadata.reflect(self.engine)


        except Exception as err:
            logger.error(err)
            sys.exit(1)

    def check_if_database_exists(self):
        return True if database_exists(self.uri) else False
            
    def create_database_if_not_exists(self, check_for_database: callable = database_exists):
        if not check_for_database(self.uri):
            while True:
                msg = input(f"The database {self.password_masked_database_uri} does not exists, would you like to create it in the destination?(y/n) ")
                if msg.lower() == "y":
                    try:
                        create_database(self.uri)
                        logger.info(f"Database {self.password_masked_database_uri} created..")
                        break
                    except Exception:
                        goodby_message(database_not_exists(self.uri), 1)
                elif msg.lower() == "n":
                    goodby_message("Destination database does not exist \nExiting ...", 0)
                print("Please, select command")

    def drop_table(self, table_name):
        with OperationContextManager(self.engine) as op:
            op.drop_table(table_name)
            logger.info(f"Table {table_name} dropped")


    def bulk_drop_tables(self, *table_name):
        try:
            for tb in table_name:
                self.drop_table(tb)
            return True
        except Exception as err:
            logger.error(err)
            sys.exit(1)

    def update_column(self, table_name, column_name, col_type, **options):            
        with OperationContextManager(self.engine) as op:
            op.alter_column(table_name, column_name,type_=col_type,postgresql_using=f"{column_name}::{col_type}") #FIXME not working

    def create_table(self, table_name: str, *columns) -> bool:
        with OperationContextManager(self.engine) as op:
            op.create_table(table_name, *columns)
            self.refresh_metadata_reflection()

        logger.info(f"Table {table_name} is created")
    
    def refresh_metadata_reflection(self):
        self.base.metadata.reflect(self.engine)


    def add_column(self, table_name: str, *column) -> bool:
        with OperationContextManager(self.engine) as op:
            for col in column:
                op.add_column(table_name, col)
            logger.info(f"Columns {column} added into table {table_name}")


    def create_fk_constraint(self, fk_constraints: list, const_columns: dict) -> bool:
        """ Get list of foreign keys from static list `fk_constraints` and create it  """
        with OperationContextManager(self.engine) as op:
            for constraint in fk_constraints:
                dest_table_name = constraint.pop("table_name")
                column_name = constraint.pop("column_name")
                source_table = constraint.pop("source_table")
                dest_column = constraint.pop("dest_column")
                temp = [i for i in const_columns[source_table]]
                if not dest_column in temp:
                    op.create_foreign_key(
                        None,
                        source_table,
                        dest_table_name,
                        [dest_column],
                        [column_name],
                        **constraint,
                    )

    def drop_fk(self, fk_constraints: str):
        with OperationContextManager(self.engine) as op:
            for fk in fk_constraints:
                op.drop_constraint(fk[1], fk[0], type_="foreignkey")
                logger.info(f"Dropped foreign key constraint {fk}")


    def db_drop_everything(self):
        """ From http://www.sqlalchemy.org/trac/wiki/UsageRecipes/DropEverything """
        tables = []
        all_foreign_keys = []
        with InspectorReflection(self.engine) as inspector: 
            for table_name in inspector.get_table_names():
                fks = []
                for fk in inspector.get_foreign_keys(table_name):
                    if not fk["name"]:
                        continue
                    fks.append(ForeignKeyConstraint((), (), name=fk["name"]))
                t = Table(table_name, self.metadata, *fks)
                tables.append(t)
                all_foreign_keys.extend(fks)
            
        with self.engine.connect() as conn:
            for foreignkey in all_foreign_keys:
                conn.execute(DropConstraint(foreignkey))

            for table in tables:
                conn.execute(DropTable(table))

            conn.commit()


    def collect_fk_and_constraint_columns(self, table_list):
        """ 
        Collect foreign key constraints for tables
        """
        dest_fk = []
        contraints_columns = []

        with InspectorReflection(self.engine) as inspector:
            for table_name in inspector.get_table_names():
                if table_name in table_list:
                    for fk in inspector.get_foreign_keys(table_name):
                        if not fk["name"]:
                            continue
                        dest_fk[fk["referred_table"]].append((table_name,fk["name"]))
                        contraints_columns[table_name].add(*fk["constrained_columns"])

        return dest_fk, contraints_columns

    def is_column_exists_in_table(self, table_name: str, column_name: str) -> bool:
        with InspectorReflection(self.engine) as inspector:
            columns = inspector.get_columns(table_name)
            for col in inspector.get_columns(table_name):
                if column_name in col["name"]:
                    return True
            return False

    def is_table_exists(self, table_name: str) -> bool:
        """Check table exist or not"""
        with InspectorReflection(self.engine) as inspector:
            tables = inspector.get_table_names()
            return table_name in tables

    def insert_data(self, table_name, data: dict):
        table = self.get_table_attribute_from_base(table_name)
        
        try:
            logger.info(f"Inserting {data} into table {table}")
            stmt = table.insert().values(**data)

        except Exception as err:
            logger.error(err, exc_info=True)
            sys.exit(1)
        self.execute_stmt(stmt=stmt)


    
    def execute_stmt(self, stmt):
        try:
            with self.engine.connect() as connection:
                connection.execute(stmt)
        except Exception as err:
            logger.error(err)

    def query_data_from_table(self, table_name, yield_per=1):
        result = self.session.query(table_name).yield_per(yield_per)
        return result


    def get_table_attribute_from_base(self, source_table_name: str):
        """
        This function gets table name attribute from sourceDB.base.classes. Example sourceDB.base.class.(table name)
        Using this attribute we can query table using sourceDB.session
        :return table attribute
        """
        try:
            return self.base.metadata.tables.get(source_table_name)
        except AttributeError as err:
            logger.error(err, exc_info=True)
            sys.exit(1)


class TableExists(Exception):
    def __init__(self, message,errors):
        self.message = message
        self.errors = errors


class FileDoesNotExists(Exception):
    def __init__(self, message):
        self.message = message


class TableDoesNotExists(Exception):
    def __init__(self, message):
        self.message = message



import logging

from sqlalchemy.dialects.mysql import (
    VARCHAR,
    INTEGER,
    NVARCHAR,
    SMALLINT,
    SET,
    BIGINT,
    BINARY,
    BOOLEAN,
    CHAR,
    DATE,
    DATETIME,
    DECIMAL,
    ENUM,
    FLOAT,
    JSON,
    NUMERIC,
    TEXT,
)
from datetime import datetime, date
from enum import Enum
from uuid import uuid4

from madmigration.config.conf import ConfigYamlManager
from madmigration.basemigration.base_migration import BaseMigrate

logger = logging.getLogger(__name__)


class MysqlMigrate(BaseMigrate): 
    def __init__(self, config: ConfigYamlManager):
        super().__init__(config)

    @staticmethod
    def get_column_type(type_name: str) -> object:
        """Get class of db type
        :param type_name: str
        :return: object class
        """
        return {
            "string": VARCHAR,
            "varchar": VARCHAR,
            "integer": INTEGER,
            "nvarchar": NVARCHAR,
            "smallint": SMALLINT,
            "set": SET,
            "bigint": BIGINT,
            "binary": BINARY,
            "boolean": BOOLEAN,
            "bool": BOOLEAN,
            "char": CHAR,
            "date": DATE,
            "datetime": DATETIME,
            "decimal": DECIMAL,
            "enum": ENUM,
            "float": FLOAT,
            "json": JSON,
            "numeric": NUMERIC,
            "text": TEXT,
            "uuid": CHAR(36),
        }.get(type_name.lower())

    
    



import click
from madmigration.controller import run
import importlib.metadata


@click.group()
def cli():
    """Controller group Cli"""


@cli.command(help="Database Migration tool")
@click.option("--file", "-f", metavar="YAML", required=True)
@click.option("--full-migrate", default=False, required=False)
def cli(file, full_migrate):
    run(config_file=file, full_migrate=full_migrate)



import yaml
import json
import os
from typing import IO, Any
from madmigration.utils.logger import configure_logging
from madmigration.config.config_schema import ConfigSchema, NoSQLConfigSchema, MigrationTablesSchema, OptionsSchema

logger = configure_logging(__file__)

class Loader(yaml.SafeLoader):
    """YAML Loader with `!import` constructor."""

    def __init__(self, stream: IO) -> None:
        """Initialise Loader."""
        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        super().__init__(stream)


def construct_include(loader: Loader, node: yaml.Node) -> Any:
    """Include file referenced at node."""

    filename = os.path.abspath(
        os.path.join(loader._root, loader.construct_scalar(node))
    )
    extension = os.path.splitext(filename)[1].lstrip(".")

    with open(filename, "r") as f:
        if extension in ("yaml", "yml"):
            return yaml.load(f, Loader)
        elif extension in ("json",):
            return json.load(f)
        else:
            return "".join(f.readlines())


yaml.add_constructor("!import", construct_include, Loader)

# Config class generates configuration based on yaml file
class ConfigYamlManager:
    def __init__(self, config_yaml):
        self.destination_mongo = False
        self.config_file = config_yaml
        self.data = self.read_config_yml_file()

        self.config_section = self.get_config_section_from_yml()
        self.source_config_section = self.get_sourceDB_config_section_from_yml()
        self.destination_config_section = self.get_destinationDB_config_section_from_yml()
        self.destination_uri = self.get_destination_db_uri()
        self.source_uri = self.get_source_db_uri()
        self.config_schema = self.verify_and_choose_ConfigSchema_based_on_database_type(self.data)
        self.migrationTables = self.config_schema.migrationTables

    def get_config_section_from_yml(self):
        return self.data.get('Configs')
    
    def get_sourceDB_config_section_from_yml(self):
        return self.config_section[0].get('SourceConfig')

    def get_destinationDB_config_section_from_yml(self):
        return self.config_section[1].get('DestinationConfig')

    def read_config_yml_file(self):
        with open(self.config_file) as f:
            return yaml.load(f, Loader=Loader)

    def get_source_db_uri(self):
        return self.source_config_section.get('dbURI')

    def get_destination_db_uri(self):
        return self.destination_config_section.get('dbURI')

    def verify_and_choose_ConfigSchema_based_on_database_type(self, data):
        if "mongodb" in self.destination_uri:
            self.destination_mongo = True
            return NoSQLConfigSchema(**data, Loader=Loader)
        else:
            return ConfigSchema(**data, Loader=Loader)

    def collect_destination_tables(self):
        """Collects all tables that the program should create"""
        table_list = []
        for migrate_table in self.migrationTables:
            table_name = migrate_table.migrationTable.DestinationTable.name
            table_list.append(table_name)
        return table_list

    def parse_migration_tables(self, tabels_schema: MigrationTablesSchema):
        try:
            source_table = tabels_schema.SourceTable
            destination_table = tabels_schema.DestinationTable
            columns = tabels_schema.MigrationColumns
            return source_table, destination_table, columns
        except Exception as err:
            logger.error(err)



import os
import sys
from pathlib import Path
from datetime import datetime, date
from enum import Enum
from uuid import uuid4, UUID


from sqlalchemy.schema import DropTable
from sqlalchemy.schema import ForeignKeyConstraint
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.engine.url import make_url

from madmigration.utils.logger import configure_logging

logger = configure_logging(__name__)

@compiles(DropTable, "postgresql")
def _compile_drop_table(element, compiler, **kwargs):
    """
    Drop table cascade with foreignkeys
    https://stackoverflow.com/questions/38678336/sqlalchemy-how-to-implement-drop-table-cascade# 
    """
    return compiler.visit_drop_table(element) + " CASCADE"


@compiles(ForeignKeyConstraint, "mysql", "mariadb")
def process(element, compiler, **kw):
    element.deferrable = element.initially = None
    return compiler.visit_foreign_key_constraint(element, **kw)


def check_file(file):
    return Path(file).is_file() and os.access(file, os.R_OK)


def file_not_found(file):
    logger.error(f"Given file does not exists file: {file}")
    sys.exit(1)


def issue_url():
    return "https://github.com/MadeByMads/mad-migration/issues"


def app_name():
    return "madmigrate"


def parse_uri(uri):
    return make_url(uri).database


def database_not_exists(database):
    """This function will be executed if there is no database exists """

    database = parse_uri(database)

    usage = [
        "",
        f"😭 Error: Source database '{database}'  does not exists",
        "",
        f"Run '{app_name()} --help' for usage.",
        "",
        f"🥳  if you think something is wrong please feel free to open issues 👉'{issue_url()}'👈 ",
        "",
        "Exiting ...",
        "",
    ]
    return "\n".join(usage)


def goodby_message(message, exit_code=0):
    print(message, flush=True)
    logger.error(message)
    sys.exit(int(exit_code))


def get_type_object(data_type):
        return {
            "UUID": UUID,
            "uuid": uuid4,
            
            "string": str,
            "str": str,
            "varchar": str,
            "text": str,
            "nvarchar": str,
            "smallint": int,
            "char": str,

            "int": int,
            "integer": int,
            "bigint": int,

            "float": float,
            "numeric": float,
            "decimal": float,

            "date": date,
            "datetime": datetime,

            "binary": bytes,
            "enum": Enum,
            "set": set,

            "json": dict,

            "boolean": bool,
            "bool": bool,

        }.get(data_type.lower())


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['madmigration',
 'madmigration.basemigration',
 'madmigration.config',
 'madmigration.db_operations',
 'madmigration.scripts',
 'madmigration.utils']

package_data = \
{'': ['*']}

install_requires = \
['alembic>=1.8.1,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'coloredlogs>=15.0.1,<16.0.0',
 'mysql>=0.0.3,<0.0.4',
 'psycopg2>=2.9.5,<3.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'pymongo>=4.3.2,<5.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'pyyaml>=6.0,<7.0',
 'regex>=2022.10.31,<2023.0.0',
 'sqlalchemy-utils>=0.38.3,<0.39.0',
 'sqlalchemy>=1.4.44,<2.0.0']

entry_points = \
{'console_scripts': ['madmigration = madmigration.main:cli']}

setup_kwargs = {
    'name': 'madmigration',
    'version': '0.1.30',
    'description': 'Mad Migration',
    'long_description': 'None',
    'author': 'Hasan',
    'author_email': 'hasan.aleeyev@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import coloredlogs

def configure_logging(file_name: str):
    logger = logging.getLogger(file_name)
    logger.setLevel(logging.DEBUG)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setLevel(logging.DEBUG)

    logger.addHandler(consoleHandler)
    fmt = '%(levelname)s file: %(name)s  func: %(funcName)20s() line: %(lineno)d  -> %(message)s'

    coloredlogs.install(level='DEBUG', logger=logger, fmt=fmt, )

    return logger



import click
from madmigration.config.conf import ConfigYamlManager
from madmigration.main import Controller, NoSQLController
from madmigration.utils.helpers import check_file, file_not_found
import importlib.metadata

__version__ = importlib.metadata.version("madmigration")


@click.group()
def cli():
    """Controller group Cli"""


@cli.command(help="Database Migration tool")

@click.option(
    "--file",
    "-f",
    metavar="YAML",
    required=True,
)

@click.version_option(__version__)

# TODO validate yaml file if given file is correct, drop fk.
def cli(file):

    if check_file(file):
        config = ConfigYamlManager(file)

        if config.destination_mongo:
            nosql = NoSQLController(config)
            nosql.run_table_migrations()
        else:
            with Controller(config) as app:
                app.run_table_migrations()
                app.run()
    else:
        file_not_found(file)

